""".. _search_methods:

Search Methods:
===================

Search methods explore the transformation space in an attempt to find a successful attack as determined by a :ref:`Goal Functions <goal_function>` and list of :ref:`Constraints <constraint>`
"""
from .search_method import SearchMethod
from .beam_search import BeamSearch
from .greedy_search import GreedySearch
from .greedy_word_swap_wir import GreedyWordSwapWIR
from .population_based_search import PopulationBasedSearch, PopulationMember
from .genetic_algorithm import GeneticAlgorithm
from .alzantot_genetic_algorithm import AlzantotGeneticAlgorithm
from .improved_genetic_algorithm import ImprovedGeneticAlgorithm
from .particle_swarm_optimization import ParticleSwarmOptimization
