""".. _goal_function:

Goal functions determine if an attack has been successful.
===========================================================
"""

from .goal_function import GoalFunction

from .classification import *
from .text import *
