"""

Goal fucntion for Classification
=======================================

"""

from .input_reduction import InputReduction
from .classification_goal_function import ClassificationGoalFunction
from .untargeted_classification import UntargetedClassification
from .targeted_classification import TargetedClassification
