"""
Universal sentence encoder
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""


from .universal_sentence_encoder import UniversalSentenceEncoder
from .multilingual_universal_sentence_encoder import (
    MultilingualUniversalSentenceEncoder,
)
