"""
Sentence Encoder
---------------------
"""


from .sentence_encoder import SentenceEncoder

from .bert import BERT
from .infer_sent import InferSent
from .thought_vector import ThoughtVector
from .universal_sentence_encoder import (
    UniversalSentenceEncoder,
    MultilingualUniversalSentenceEncoder,
)
