"""

Goal Function for Text to Text case
=======================================

"""

from .minimize_bleu import MinimizeBleu
from .non_overlapping_output import NonOverlappingOutput
from .text_to_text_goal_function import TextToTextGoalFunction
